# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo
from mcpi.vec3 import Vec3

from .thing import Thing


class Wall(Thing):
    height = 5
    length = 10
    width = 2

    def create(self):
        self.set_blocks(
            Vec3(self.position.x, self.position.y, self.position.z),
            Vec3(self.position.x + self.length - 1,
                 self.position.y + self.height - 1,
                 self.position.z + self.width - 1),
            self.block)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo


class Renderer:
    """ Base class for all McThings Renderers """

    """
    Base class for all McThings Renderers
    
    A renderer reads the blocks data from mcthings.core.BlocksMemory and render it
    with a specific engine. For example, Raspberry PI uses the Python API to do it.
    """

    def render(self, memory_chunk):
        """
        Render the blocks included in the memory_chunk at position in the world

        :param memory_chunk: chunk of blocks to be rendered
        :return:
        """
        pass



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi
from mcpi.vec3 import Vec3

from .thing import Thing


class Fence(Thing):
    """
    Build a block plane and empty it to create the fence
    """

    fence_space = 5
    """ Space between the fence and the thing fenced """
    thick = 1
    height = None
    thing = None

    def create(self):
        """
        Create a fence around the configured thing
        :return:
        """
        if self.thing is None:
            raise RuntimeError("Thing to be fenced is not defined")

        self.add_child(self.thing)

        init_x = self.thing.position.x - self.fence_space - self.thick
        init_y = self.thing.position.y
        init_z = self.thing.position.z - self.fence_space - self.thick

        self._position = Vec3(init_x, init_y, init_z)

        end_x = self.thing.end_position.x + self.fence_space + self.thick
        end_y = self.thing.end_position.y
        if self.height:
            end_y = self.thing.position.y + (self.height - 1)
        end_z = self.thing.end_position.z + self.fence_space + self.thick

        self._end_position = Vec3(end_x, end_y, end_z)

        self.set_blocks(
            Vec3(init_x, init_y, init_z),
            Vec3(end_x, end_y, end_z),
            self.block)

        # Fill the prism with air to became a rectangular wall
        self.set_blocks(
            Vec3(init_x + self.thick, init_y, init_z + self.thick),
            Vec3(end_x - self.thick, end_y, end_z - self.thick),
            mcpi.block.AIR)

        # Rebuild the thing because it is destroyed when emptying the fence
        # if we are not removing the fence
        if self.block != mcpi.block.AIR:
            self.thing.create()



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

# TODO: at some point this must be a real Singleton
import pickle

from mcpi.vec3 import Vec3

from mcthings.utils import build_schematic_nbt
from mcthings.world import World


class Scene:
    """
    A scene is a container for all the things built using McThings.
    A scene can be built, unbuilt and moved. There is only one scene
    in a program using McThings. Things built are added automatically to
    the Scene. A Scene can also be loaded from a file, and
    it can be saved to a file.

    Before adding Things to the Scene, it must be connected to a
    Minecraft server (fill the Scene.server attribute)
    """

    def __init__(self, renderer):
        self.things = []
        """ map with the things in the scene """
        self._decorators = []
        """ decorators for the scene """
        self._position = None
        """ position in the world of the scene """
        self._end_position = None
        """ end position in the world of the scene """
        self._renderer = renderer

        World.add_scene(self)

    @property
    def end_position(self):
        """ end position of the thing """
        return self._end_position

    @property
    def position(self):
        """ initial position of the thing """
        return self._position

    def add(self, thing):
        """ Add a new thing to the scene """
        if not self.things:
            # The initial position of the scene is the position
            # of its first thing added
            self._position = thing.position
        self.things.append(thing)

    def add_decorator(self, decorator):
        """ Add a new decorator to the scene """
        self._decorators.append(decorator)

    def decorate(self):
        """
        Call all decorators for the current Scene

        :return:
        """

        for decorator in self._decorators:
            decorator(self).decorate()

    def build(self):
        """ Build all the things inside the Scene """
        for thing in self.things:
            thing.build()

        (min_pos, max_pos) = self.find_bounding_box()
        self._end_position = max_pos

    def unbuild(self):
        """ Unbuild all the things inside the Scene """
        for thing in self.things:
            thing.unbuild()

    def reposition(self, position):
        """
        Move all the things in the scene to a new relative position

        :param position: new position for the Scene
        :return:
        """

        # All the things inside the scene must be moved
        diff_x = position.x - self._position.x
        diff_y = position.y - self._position.y
        diff_z = position.z - self._position.z

        for thing in self.things:
            repos_x = thing.position.x + diff_x
            repos_y = thing.position.y + diff_y
            repos_z = thing.position.z + diff_z

            thing._position = (Vec3(repos_x, repos_y, repos_z))

    def move(self, position):
        """
        Move the scene to a new position

        :param position: new position
        :return:
        """

        self.unbuild()
        self.reposition(position)
        self.build()

    def load(self, file_path):
        """ Load a scene from a file (but no build it yet) """
        self.things = pickle.load(open(file_path, "rb"))
        if self.things:
            self._position = self.things[0].position

    def save(self, file_path):
        """ Save a scene to a file """
        pickle.dump(self.things, open(file_path, "wb"))

    def find_bounding_box(self):
        """ Compute the bounding box of the Scene """

        def update_box(box_pos_min, box_pos_max, pos):
            # Update box_pos_min and box_pos_max checking pos
            box_pos_min.x = pos.x if pos.x < box_pos_min.x else box_pos_min.x
            box_pos_min.y = pos.y if pos.y < box_pos_min.y else box_pos_min.y
            box_pos_min.z = pos.z if pos.z < box_pos_min.z else box_pos_min.z
            box_pos_max.x = pos.x if pos.x > box_pos_max.x else box_pos_max.x
            box_pos_max.y = pos.y if pos.y > box_pos_max.y else box_pos_max.y
            box_pos_max.z = pos.z if pos.z > box_pos_max.z else box_pos_max.z

            return box_pos_min, box_pos_max

        # Default init values
        min_pos = Vec3(self._position.x, self._position.y, self._position.z)
        max_pos = Vec3(self._position.x, self._position.y, self._position.z)

        # Find the bounding box for the scene
        for thing in self.things:
            min_pos, max_pos = update_box(min_pos, max_pos, thing.position)
            if thing.end_position:
                min_pos, max_pos = update_box(min_pos, max_pos, thing.end_position)

        return min_pos, max_pos

    def to_schematic(self, file_path, block_data=False):
        """
        Save the Scene into a Schematic file

        :param file_path: file in which to export the Scene in Schematic format
        :param block_data: extract blocks ids and data (much slower)
        :return: the Schematic object
        """

        (min_pos, max_pos) = self.find_bounding_box()

        build_schematic_nbt(min_pos, max_pos, block_data).write_file(file_path)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi

from mcpi.vec3 import Vec3

from .thing import Thing


class BlocksGallery(Thing):
    # https://www.minecraftinfo.com/idlist.htm
    MAX_BLOCK_NUMBER = 247

    def create(self):
        """
        Show all possible block types in a line
        :return:
        """

        for i in range(1, self.MAX_BLOCK_NUMBER):
            p = self.position
            self.set_block(Vec3(p.x + i, p.y, p.z), mcpi.block.Block(i))

        self._end_position = Vec3(p.x + self.MAX_BLOCK_NUMBER - 1, p.y, p.z)



# TODO: at some point this must be a real Singleton

from mcpi.minecraft import Minecraft
# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi.block

from mcthings.blocks_memory import BlocksMemory


class Decorator:
    """
    A Decorator is able to decorate a Thing based on its characteristics.

    If a Thing has decorators, they will be called after the build of the Thing.
    """

    block = mcpi.block.AIR
    """ Base block for the decorator """

    def __init__(self, thing):
        self._blocks_memory = BlocksMemory()
        self._renderer = thing._renderer
        self._thing = thing

    def create(self):
        """
        Create the decorator

        :return:
        """

    def set_block(self, pos, block, data=None):
        self._blocks_memory.set_block(pos, block, data)

    def set_blocks(self, init_pos, end_pos, block):
        """ Add a cuboid with the same block for all blocks and without specific data"""
        self._blocks_memory.set_blocks(init_pos, end_pos, block)

    def render(self):
        """
        Renders the decorator

        :return:
        """
        self._renderer.render(self._blocks_memory)

    def decorate(self):
        """
        Decorate the thing

        :return:
        """

        self.create()
        self.render()
